@@ -1,10 +1,13 @@
+import traceback
 import logging
 import psycopg2
 import psycopg2.extras
 import psycopg2.extensions as ext
 import sqlparse
-from .packages import pgspecial
-from .encodingutils import unicode2utf8
+import pgspecial as special
+from .packages.function_metadata import FunctionMetadata
+from .encodingutils import unicode2utf8, PY2
+
 
 _logger = logging.getLogger(__name__)
 
@@ -49,6 +52,7 @@ def register_json_typecasters(conn, loads_fn):
 
     return available
 
+
 def register_hstore_typecaster(conn):
     """
     Instead of using register_hstore() which converts hstore into a python
@@ -64,16 +68,17 @@ def register_hstore_typecaster(conn):
         except Exception:
             pass
 
+
 class PGExecute(object):
 
+    # The boolean argument to the current_schemas function indicates whether
+    # implicit schemas, e.g. pg_catalog
     search_path_query = '''
-        SELECT * FROM unnest(current_schemas(false))'''
+        SELECT * FROM unnest(current_schemas(true))'''
 
     schemata_query = '''
         SELECT  nspname
         FROM    pg_catalog.pg_namespace
-        WHERE   nspname !~ '^pg_'
-                AND nspname <> 'information_schema'
         ORDER BY 1 '''
 
     tables_query = '''
@@ -82,9 +87,7 @@ class PGExecute(object):
         FROM 	pg_catalog.pg_class c
                 LEFT JOIN pg_catalog.pg_namespace n
                     ON n.oid = c.relnamespace
-        WHERE 	c.relkind IN ('r','v', 'm') -- table, view, materialized view
-                AND n.nspname !~ '^pg_'
-                AND nspname <> 'information_schema'
+        WHERE 	c.relkind = ANY(%s)
         ORDER BY 1,2;'''
 
     columns_query = '''
@@ -96,21 +99,22 @@ class PGExecute(object):
                     ON att.attrelid = cls.oid
                 INNER JOIN pg_catalog.pg_namespace nsp
                     ON cls.relnamespace = nsp.oid
-        WHERE 	cls.relkind IN ('r', 'v', 'm')
-                AND nsp.nspname !~ '^pg_'
-                AND nsp.nspname <> 'information_schema'
+        WHERE 	cls.relkind = ANY(%s)
                 AND NOT att.attisdropped
                 AND att.attnum  > 0
         ORDER BY 1, 2, 3'''
 
     functions_query = '''
-        SELECT 	DISTINCT  --multiple dispatch means possible duplicates
-                n.nspname schema_name,
-                p.proname func_name
+        SELECT 	n.nspname schema_name,
+                p.proname func_name,
+                pg_catalog.pg_get_function_arguments(p.oid) arg_list,
+                pg_catalog.pg_get_function_result(p.oid) return_type,
+                p.proisagg is_aggregate,
+                p.proiswindow is_window,
+                p.proretset is_set_returning
         FROM 	pg_catalog.pg_proc p
                 INNER JOIN pg_catalog.pg_namespace n
                     ON n.oid = p.pronamespace
-        WHERE 	n.nspname NOT IN ('pg_catalog', 'information_schema')
         ORDER BY 1, 2'''
 
 
@@ -123,31 +127,87 @@ class PGExecute(object):
     FROM pg_catalog.pg_database d
     ORDER BY 1;"""
 
-    def __init__(self, database, user, password, host, port):
+    datatypes_query = '''
+        SELECT n.nspname schema_name,
+               t.typname type_name
+        FROM   pg_catalog.pg_type t
+               INNER JOIN pg_catalog.pg_namespace n
+                  ON n.oid = t.typnamespace
+        WHERE ( t.typrelid = 0  -- non-composite types
+                OR (  -- composite type, but not a table
+                      SELECT c.relkind = 'c'
+                      FROM pg_catalog.pg_class c
+                      WHERE c.oid = t.typrelid
+                    )
+              )
+              AND NOT EXISTS( -- ignore array types
+                    SELECT  1
+                    FROM    pg_catalog.pg_type el
+                    WHERE   el.oid = t.typelem AND el.typarray = t.oid
+                  )
+              AND n.nspname <> 'pg_catalog'
+              AND n.nspname <> 'information_schema'
+        ORDER BY 1, 2;'''
+
+    def __init__(self, database, user, password, host, port, dsn):
         self.dbname = database
         self.user = user
         self.password = password
         self.host = host
         self.port = port
+        self.dsn = dsn
         self.connect()
 
     def connect(self, database=None, user=None, password=None, host=None,
-            port=None):
+            port=None, dsn=None):
 
-        db = unicode2utf8(database or self.dbname)
-        user = unicode2utf8(user or self.user)
-        password = unicode2utf8(password or self.password)
-        host = unicode2utf8(host or self.host)
-        port = unicode2utf8(port or self.port)
-        conn = psycopg2.connect(database=db, user=user, password=password,
-                host=host, port=port)
+        db = (database or self.dbname)
+        user = (user or self.user)
+        password = (password or self.password)
+        host = (host or self.host)
+        port = (port or self.port)
+        dsn = (dsn or self.dsn)
+        if dsn:
+            if password:
+                dsn = "{0} password={1}".format(dsn, password)
+            conn = psycopg2.connect(dsn=unicode2utf8(dsn))
+            cursor = conn.cursor()
+            # When we connect using a DSN, we don't really know what db,
+            # user, etc. we connected to. Let's read it.
+            db = self._select_one(cursor, 'select current_database()')
+            user = self._select_one(cursor, 'select current_user')
+            host = self._select_one(cursor, 'select inet_server_addr()')
+            port = self._select_one(cursor, 'select inet_server_port()')
+        else:
+            conn = psycopg2.connect(
+                    database=unicode2utf8(db),
+                    user=unicode2utf8(user),
+                    password=unicode2utf8(password),
+                    host=unicode2utf8(host),
+                    port=unicode2utf8(port))
+        conn.set_client_encoding('utf8')
         if hasattr(self, 'conn'):
             self.conn.close()
         self.conn = conn
         self.conn.autocommit = True
+        self.dbname = db
+        self.user = user
+        self.password = password
+        self.host = host
+        self.port = port
         register_json_typecasters(self.conn, self._json_typecaster)
         register_hstore_typecaster(self.conn)
 
+    def _select_one(self, cur, sql):
+        """
+        Helper method to run a select and retrieve a single field value
+        :param cur: cursor
+        :param sql: string
+        :return: string
+        """
+        cur.execute(sql)
+        return cur.fetchone()
+
     def _json_typecaster(self, json_data):
         """Interpret incoming JSON data as a string.
 
@@ -157,63 +217,87 @@ class PGExecute(object):
         See http://initd.org/psycopg/docs/connection.html#connection.encoding
         """
 
-        return json_data.decode(self.conn.encoding)
+        if PY2:
+            return json_data.decode(self.conn.encoding)
+        else:
+            return json_data
 
-    def run(self, statement):
-        """Execute the sql in the database and return the results. The results
-        are a list of tuples. Each tuple has 3 values (rows, headers, status).
+    def run(self, statement, pgspecial=None, exception_formatter=None,
+            on_error_resume=False):
+        """Execute the sql in the database and return the results.
+
+        :param statement: A string containing one or more sql statements
+        :param pgspecial: PGSpecial object
+        :param exception_formatter: A callable that accepts an Exception and
+               returns a formatted (title, rows, headers, status) tuple that can
+               act as a query result. If an exception_formatter is not supplied,
+               psycopg2 exceptions are always raised.
+        :param on_error_resume: Bool. If true, queries following an exception
+               (assuming exception_formatter has been supplied) continue to
+               execute.
+
+        :return: Generator yielding tuples containing
+                 (title, rows, headers, status, query, success)
         """
 
         # Remove spaces and EOL
         statement = statement.strip()
         if not statement:  # Empty string
-            yield (None, None, None, None)
+            yield (None, None, None, None, statement, False)
 
         # Split the sql into separate queries and run each one.
         for sql in sqlparse.split(statement):
             # Remove spaces, eol and semi-colons.
             sql = sql.rstrip(';')
 
-            # Check if the command is a \c or 'use'. This is a special
-            # exception that cannot be offloaded to `pgspecial` lib. Because we
-            # have to change the database connection that we're connected to.
-            if sql.startswith('\c') or sql.lower().startswith('use'):
-                _logger.debug('Database change command detected.')
-                try:
-                    dbname = sql.split()[1]
-                except:
-                    _logger.debug('Database name missing.')
-                    raise RuntimeError('Database name missing.')
-                self.connect(database=dbname)
-                self.dbname = dbname
-                _logger.debug('Successfully switched to DB: %r', dbname)
-                yield (None, None, None, 'You are now connected to database "%s" as '
-                        'user "%s"' % (self.dbname, self.user))
-            else:
-                try:   # Special command
+            try:
+                if pgspecial:
+                    # First try to run each query as special
                     _logger.debug('Trying a pgspecial command. sql: %r', sql)
                     cur = self.conn.cursor()
-                    for result in pgspecial.execute(cur, sql):
-                        yield result
-                except KeyError:  # Regular SQL
-                    yield self.execute_normal_sql(sql)
+                    try:
+                        for result in pgspecial.execute(cur, sql):
+                            yield result + (sql, True)
+                        continue
+                    except special.CommandNotFound:
+                        pass
+
+                # Not a special command, so execute as normal sql
+                yield self.execute_normal_sql(sql) + (sql, True)
+            except psycopg2.DatabaseError as e:
+                _logger.error("sql: %r, error: %r", sql, e)
+                _logger.error("traceback: %r", traceback.format_exc())
+
+                if (isinstance(e, psycopg2.OperationalError)
+                        or not exception_formatter):
+                    # Always raise operational errors, regardless of on_error
+                    # specification
+                    raise
+
+                yield None, None, None, exception_formatter(e), sql, False
+
+                if not on_error_resume:
+                    break
 
     def execute_normal_sql(self, split_sql):
+        """Returns tuple (title, rows, headers, status)"""
         _logger.debug('Regular sql statement. sql: %r', split_sql)
         cur = self.conn.cursor()
         cur.execute(split_sql)
+
         try:
             title = self.conn.notices.pop()
         except IndexError:
             title = None
+
         # cur.description will be None for operations that do not return
         # rows.
         if cur.description:
             headers = [x[0] for x in cur.description]
-            return (title, cur, headers, cur.statusmessage)
+            return title, cur, headers, cur.statusmessage
         else:
             _logger.debug('No rows in result.')
-            return (title, None, None, cur.statusmessage)
+            return title, None, None, cur.statusmessage
 
     def search_path(self):
         """Returns the current search path as a list of schema names"""
@@ -231,21 +315,60 @@ class PGExecute(object):
             cur.execute(self.schemata_query)
             return [x[0] for x in cur.fetchall()]
 
+    def _relations(self, kinds=('r', 'v', 'm')):
+        """Get table or view name metadata
+
+        :param kinds: list of postgres relkind filters:
+                'r' - table
+                'v' - view
+                'm' - materialized view
+        :return: (schema_name, rel_name) tuples
+        """
+
+        with self.conn.cursor() as cur:
+            sql = cur.mogrify(self.tables_query, [kinds])
+            _logger.debug('Tables Query. sql: %r', sql)
+            cur.execute(sql)
+            for row in cur:
+                yield row
+
     def tables(self):
-        """Returns a list of (schema_name, table_name) tuples """
+        """Yields (schema_name, table_name) tuples"""
+        for row in self._relations(kinds=['r']):
+            yield row
+
+    def views(self):
+        """Yields (schema_name, view_name) tuples.
+
+            Includes both views and and materialized views
+        """
+        for row in self._relations(kinds=['v', 'm']):
+            yield row
+
+    def _columns(self, kinds=('r', 'v', 'm')):
+        """Get column metadata for tables and views
+
+        :param kinds: kinds: list of postgres relkind filters:
+                'r' - table
+                'v' - view
+                'm' - materialized view
+        :return: list of (schema_name, relation_name, column_name) tuples
+        """
 
         with self.conn.cursor() as cur:
-            _logger.debug('Tables Query. sql: %r', self.tables_query)
-            cur.execute(self.tables_query)
-            return cur.fetchall()
+            sql = cur.mogrify(self.columns_query, [kinds])
+            _logger.debug('Columns Query. sql: %r', sql)
+            cur.execute(sql)
+            for row in cur:
+                yield row
 
-    def columns(self):
-        """Returns a list of (schema_name, table_name, column_name) tuples"""
+    def table_columns(self):
+        for row in self._columns(kinds=['r']):
+            yield row
 
-        with self.conn.cursor() as cur:
-            _logger.debug('Columns Query. sql: %r', self.columns_query)
-            cur.execute(self.columns_query)
-            return cur.fetchall()
+    def view_columns(self):
+        for row in self._columns(kinds=['v', 'm']):
+            yield row
 
     def databases(self):
         with self.conn.cursor() as cur:
@@ -254,10 +377,19 @@ class PGExecute(object):
             return [x[0] for x in cur.fetchall()]
 
     def functions(self):
-        """Yields tuples of (schema_name, function_name)"""
+        """Yields FunctionMetadata named tuples"""
 
         with self.conn.cursor() as cur:
             _logger.debug('Functions Query. sql: %r', self.functions_query)
             cur.execute(self.functions_query)
+            for row in cur:
+                yield FunctionMetadata(*row)
+
+    def datatypes(self):
+        """Yields tuples of (schema_name, type_name)"""
+
+        with self.conn.cursor() as cur:
+            _logger.debug('Datatypes Query. sql: %r', self.datatypes_query)
+            cur.execute(self.datatypes_query)
             for row in cur:
                 yield row
